﻿#region Header

// The sole owner of this code is Dundas Data Visualization Inc.
// Copyright (c) 2015 Dundas Data Visualization Inc.
// All rights reserved.

#endregion Header

using Dundas.BI.Data.Parameters;
using Dundas.BI.Data.Transforms;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net.Mime;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace Dundas.BI.Sample.CustomTransformSample
{
    /// <summary>
    /// This class represents a data boundary transform.
    /// </summary>
    public class DataBoundaryTransform : Transform
	{
		#region Static Fields

        /// <summary>
        /// The upper boundary setting identifier.
        /// </summary>
        private static readonly Guid BOUNDARY_VIOLATION_ACTION_ID = new Guid("d4c68615-2ada-4923-8d75-d900711a4293");

		/// <summary>
		/// The upper boundary setting identifier.
		/// </summary>
		private static readonly Guid UPPER_BOUNDARY_ID = new Guid("b60311fd-b3b9-401a-b61d-2fa078c4a05d");

        /// <summary>
        /// The use upper boundary setting identifier.
        /// </summary>
        private static readonly Guid USE_UPPER_BOUNDARY_ID = new Guid("118ae60c-90a5-4693-a24a-5308877811a8");

        /// <summary>
        /// The lower boundary setting identifier.
        /// </summary>
        private static readonly Guid LOWER_BOUNDARY_ID = new Guid("a60efc80-41fa-451b-875d-50c8bbfa1cb7");

        /// <summary>
        /// The use lower boundary setting identifier.
        /// </summary>
        private static readonly Guid USE_LOWER_BOUNDARY_ID = new Guid("c9306ec6-4dcc-4db6-af5a-8e5186345a5e");

		/// <summary>
		/// The columns to apply to setting identifier.
		/// </summary>
		private static readonly Guid COLUMNS_TO_APPLY_TO = new Guid("ee504e77-a6a1-423e-9ff9-1eb82af4c565");

		#endregion Static Fields

		#region Fields

		private List<string> _columnNamesToApplyTo;

		#endregion Fields

		#region Private Properties

		/// <summary>
		/// Gets the column names to apply to.
		/// </summary>
		/// <value>
		/// The column names to apply to.
		/// </value>
		private List<string> ColumnNamesToApplyTo
		{
			get
			{
				if (_columnNamesToApplyTo == null)
				{
                    _columnNamesToApplyTo = new List<string>();

                    for (int index = 5; index < this.TransformSettings.Count; index++)
                    {
                        SingleBooleanValue singleBooleanValue =
                            this.GetConfigurationValue(this.TransformSettings[index]) as SingleBooleanValue;

                        if(singleBooleanValue.Value)
                        {
                            _columnNamesToApplyTo.Add(this.TransformSettings[index].Name);
                        }
                    }
                    
				}

				return _columnNamesToApplyTo;
			}
		}

		#endregion Private Properties

		#region Public Properties

        /// <summary>
        /// Gets the standard component description.
        /// </summary>
        public override string ComponentDescription
        {
	        get { return "This transform will enforce a boundary on the data."; }
        }

        /// <summary>
        /// Gets the component ID.
        /// </summary>
        public override Guid ComponentId
        {
	        get { return new Guid("3fa26810-cbae-450c-ae53-986a228cdcd9"); }
        }

        /// <summary>
        /// Gets the standard component name.
        /// </summary>
        public override string ComponentName
        {
            get { return "Data Boundary"; }
        }

		/// <summary>
		/// Gets the type of the processing.
		/// </summary>
		public override ProcessBehaviors ProcessingType
		{
			get
			{
				return ProcessBehaviors.Step | ProcessBehaviors.SingleRecord | ProcessBehaviors.SingleElement;
			}
		}

        /// <summary>
        /// Gets a value indicating whether this transform is using a custom configuration UI.
        /// </summary>
        public override bool HasCustomConfigurationUI
        {
            get
            {
                return true;
            }
        }

		#endregion Public Properties

		#region Private Methods

        /// <summary>
        /// Applies the boundary to source value.
        /// </summary>
        /// <param name="sourceValue">The source value.</param>
        /// <returns>If the value is ouside the boundary then, otherwise returns the original source value.</returns>
		private object ApplyBoundaryToSourceValue(object sourceValue)
		{
			SingleNumberValue upperParameter = this.GetConfigurationValue(UPPER_BOUNDARY_ID) as SingleNumberValue;
            SingleNumberValue lowerParameter = this.GetConfigurationValue(LOWER_BOUNDARY_ID) as SingleNumberValue;
            SingleBooleanValue useUpperParameter = this.GetConfigurationValue(USE_UPPER_BOUNDARY_ID) as SingleBooleanValue;
            SingleBooleanValue useLowerParameter = this.GetConfigurationValue(USE_LOWER_BOUNDARY_ID) as SingleBooleanValue;
            SingleNumberValue boundaryActionId = this.GetConfigurationValue(BOUNDARY_VIOLATION_ACTION_ID) as SingleNumberValue;

			int upperValue = 
                (int)upperParameter.Value;

            int lowerValue = 
                (int)lowerParameter.Value;

            bool useUpperBoundary =
                (bool)useUpperParameter.Value;

            bool useLowerBoundary =
                (bool)useLowerParameter.Value;

            Dundas.BI.Number? number = 
                Dundas.BI.Number.ConvertToNumber(sourceValue);

            if (useUpperBoundary)
            {
                if (number > upperValue)
                {
                    object returnObject = null;

                    switch ((int)boundaryActionId.Value)
                    {
                        case 0:
                            returnObject = upperValue;
                            break;
                        case 1:
                            returnObject = 0;
                            break;
                        default:
                            returnObject = null;
                            break;
                    }

                    return
                        returnObject;
                }

            }

            if (useLowerBoundary)
            {

                if (number < lowerValue)
                {
                    object returnObject = null;

                    switch ((int)boundaryActionId.Value)
                    {
                        case 0:
                            returnObject = lowerValue;
                            break;
                        case 1:
                            returnObject = 0;
                            break;
                        default:
                            returnObject = null;
                            break;
                    }

                    return
                        returnObject;
                }

            }

			return sourceValue;
		}

        /// <summary>
        /// Defines the inputs.
        /// </summary>
        /// <returns>A list of InputConnectors.</returns>
		private IList<InputConnector> DefineInputs()
		{
			List<InputConnector> inputs = new List<InputConnector>();

			InputConnector input =
				new InputConnector(
					this,
					new Guid("cf45944f-57bd-4418-b534-6dc1042e32cc"),
					"Data Boundary Input",
					"Data Boundary Input Description",
					ConnectorDataFormats.Tabular,
					null,
					null,
					null,
					false,
					ErrorHandlingPolicy.Fail
					);
			inputs.Add(input);



			return inputs;
		}

        /// <summary>
        /// Defines the output.
        /// </summary>
        /// <returns>A list of OutputConnectors.</returns>
        private OutputConnector DefineOutput()
        {
	        OutputConnector output =
		        new OutputConnector(
			        this,
			        new Guid("eb297a8f-87e3-496d-88a9-8f30c9289120"),
                    "Data Boundary Output",
                    "Data Boundary  Output Description",
			        ConnectorDataFormats.Tabular
			        );

	        return output;
        }


        /// <summary>
        /// Defines the settings.
        /// </summary>
        /// <returns>A list of component settings.</returns>
		private IList<ComponentSetting> DefineSettings()
		{
			List<ComponentSetting> settings = new List<ComponentSetting>();

            NumericSetting boundaryViolationActionSetting = 
                new NumericSetting(
                    BOUNDARY_VIOLATION_ACTION_ID,
                    "Boundary Violation Action",
                    "What to do if the boundary is violated.",
                    false,
                    0,
                    ValidValuesSource.ManualList,
                    false
                );

            boundaryViolationActionSetting.AddValidValue(
                0, 
                "Boundary", 
                "Use boundary as replacement."
                );
            boundaryViolationActionSetting.AddValidValue(
                1, 
                "Zero", 
                "Use zero as replacement."
                );
            boundaryViolationActionSetting.AddValidValue(
                2, 
                "Null", 
                "Use null as replacement."
                );

            settings.Add(boundaryViolationActionSetting);

            BooleanSetting useUpperBoundarySetting =
                new BooleanSetting(
                    USE_UPPER_BOUNDARY_ID,
                    "Use Upper Boundary",
                    "Will enforce an upper boundary",
                    false,
                    false,
                    false
                    );

            settings.Add(useUpperBoundarySetting);

            NumericSetting upperBoundarySetting =
	            new NumericSetting(
		            UPPER_BOUNDARY_ID,
		            "Upper Boundary",
		            "The upper boundary",
		            false,
		            0,
		            int.MinValue,
		            int.MaxValue,
		            false
		            );


			settings.Add(upperBoundarySetting);

            BooleanSetting useLowerBoundarySetting =
                new BooleanSetting(
                    USE_LOWER_BOUNDARY_ID,
                    "Use Lower Boundary",
                    "Will enforce an lower boundary",
                    false,
                    true,
                    false
                    );

            settings.Add(useLowerBoundarySetting);

            NumericSetting lowerBoundarySetting =
                new NumericSetting(
                    LOWER_BOUNDARY_ID,
                    "Lower Boundary",
                    "The lower boundary",
                    false,
                    0,
                    int.MinValue,
                    int.MaxValue,
                    false
                    );


            settings.Add(lowerBoundarySetting);


			return settings;
		}

		#endregion Private Methods

		#region Protected Methods

		/// <summary>
		/// Populates the process transform metadata by describing the data inputs,   outputs and any available settings.
		/// </summary>
		/// <param name="inputs">The inputs.</param>
		/// <param name="output">The output.</param>
		/// <param name="settings">The transform settings.</param>
		protected override void GetMetadata(out IList<InputConnector> inputs, out OutputConnector output, out IList<Dundas.BI.Data.Parameters.ComponentSetting> settings)
		{
			inputs = this.DefineInputs();
			output = this.DefineOutput();
			settings = this.DefineSettings();
		}

		#endregion Protected Methods

		#region Public Methods

        /// <summary>
        /// Gets the custom configuration UI based on the requested content type.
        /// </summary>
        /// <param name="contentType">The content type that the UI is requested for.</param>
        /// <returns>
        /// The string holding the custom UI, or <see langword="null" /> if none is supported for the content type.
        /// </returns>
        public override string GetCustomConfigurationUI(System.Net.Mime.ContentType contentType)
        {
            if(contentType == null)
            {
                return null;
            }

            if(contentType.Equals(MediaTypeNames.Text.Html))
            {
                Assembly assembly = typeof(DataBoundaryTransform).Assembly;

                using (StreamReader streamReader =
                    new StreamReader(assembly.GetManifestResourceStream(
                        "Dundas.BI.Sample.CustomTransformSample.DataBoundaryUI.html"
                        )
                    )
                )
                {
                    string contents = streamReader.ReadToEnd();

                    return contents;
                }
            }

            return null;
            
        }

        /// <summary>
        /// Fetches the value for the specified output element for run time data retrieval.
        /// </summary>
        /// <param name="outputElement">The output element.</param>
        /// <param name="skipValidation">If set to <see langword="true" /> parameter validation is skipped for performance reasons.</param>
        /// <returns>
        /// The element value.
        /// </returns>
        /// <exception cref="System.ArgumentNullException">outputElement</exception>
        /// <exception cref="System.ArgumentException"></exception>
        public override object FetchElementValue(ConnectorElement outputElement, bool skipValidation)
        {
	        if (!skipValidation)
	        {
		        if (outputElement == null) { throw new ArgumentNullException("outputElement"); }
		        if (outputElement.Connector.Transform != this) { throw new ArgumentException(this.Localize("GS_EX_Invalid_ParentElement")); }
	        }

	        ConnectorElement inputElement = null;

	        if (outputElement.UpstreamElement == null)
	        {
		        inputElement = this.FindConnectorElement(outputElement.Dependencies.First());
	        }
	        else
	        {
		        inputElement = outputElement.UpstreamElement;
	        }

	        object sourceValue = this.Input.ConnectedTo.DataReader[inputElement.UpstreamElement.Id];

	        if (!ColumnNamesToApplyTo.Contains(inputElement.UpstreamElement.NativeElementName))
	        {
		        return sourceValue;
	        }
	        else
	        {
		        return ApplyBoundaryToSourceValue(sourceValue);
	        }
        }

		/// <summary>
		/// Generates the list of <see cref="T:Dundas.BI.Data.Transforms.OutputElementSource" />, which are used elements for an <see cref="T:Dundas.BI.Data.Transforms.OutputConnector" />.
		/// </summary>
		/// <returns>
		/// The list of output element sources.
		/// </returns>
		public override IList<OutputElementSource> GenerateOutputElementSources()
		{
			List<OutputElementSource> outputSources = new List<OutputElementSource>();

			foreach (var element in this.Input.Elements)
			{
                BooleanSetting useColumnSetting =
                    new BooleanSetting(
                        Guid.NewGuid(),
                        element.NativeElementName,
                        string.Format(
                            CultureInfo.CurrentCulture,
                            "Use {0} element",
                            element.NativeElementName
                            ),
                        false,
                        false,
                        false
                        );
                this.AppendTransformSettings(useColumnSetting);

				outputSources.Add(new OutputElementSource(this.Output, element, true, false));

			}
            
			return outputSources;
		}

        /// <summary>
        /// Reads a single record for the transform output, if there are any. If there are no more output records, it returns <see langword="false" />.
        /// The base class implementation calls Read on all the input readers and return <see langword="true" /> if at least one of these operations is successful.
        /// Transforms that process a single record at one time do not need to override the base class implementation.
        /// Transforms that process multiple records at one time need to override the base class implementation by retrieving the necessary input data and also performing output record processing.
        /// If the Read operation is successful, the transform is responsible for providing corresponding data values on subsequent FetchElementValue calls.
        /// </summary>
        /// <returns>
        ///   <see langowrd="true" /> if the read returns records, otherwise <see langword="false" />.
        /// </returns>
		public override bool Read()
		{
			bool readValue = base.Read();

			// Reset the column names to read setting, after last read.
			if(!readValue)
			{
				_columnNamesToApplyTo = null;
			}

			return readValue;
		}

		#endregion Public Methods
	}
}